from typing import overload, Union

from pykotor.common.stream import BinaryReader, BinaryWriter
from pykotor.resource.formats.twoda import TwoDA, TwoDABinaryReader, TwoDABinaryWriter, TwoDACSVWriter, TwoDACSVReader
from pykotor.resource.type import FileFormat, SOURCE_TYPES, TARGET_TYPES


@overload
def detect_2da(source: str, offset: int = 0):
    ...


@overload
def detect_2da(source: bytes, offset: int = 0):
    ...


@overload
def detect_2da(source: bytearray, offset: int = 0):
    ...


@overload
def detect_2da(source: BinaryReader, offset: int = 0):
    ...


def detect_2da(source: SOURCE_TYPES, offset: int = 0) -> FileFormat:
    """
    Returns what format the TwoDA data is believed to be in. This function performs a basic check and does not guarantee
    accuracy of the result or integrity of the data.

    Args:
        source: Source of the TwoDA data.
        offset: Offset into the data.

    Returns:
        The format of the TwoDA data.
    """
    try:
        if isinstance(source, str):
            with BinaryReader.from_file(source, offset) as reader:
                file_format = FileFormat.BINARY if reader.read_string(4) == "2DA " else FileFormat.CSV
        elif isinstance(source, bytes) or isinstance(source, bytearray):
            file_format = FileFormat.BINARY if source[:4].decode('ascii', 'ignore') == "2DA " else FileFormat.CSV
        elif isinstance(source, BinaryReader):
            file_format = FileFormat.BINARY if source.read_string(4) == "2DA " else FileFormat.CSV
            source.skip(-4)
        else:
            file_format = FileFormat.INVALID
    except IOError:
        file_format = FileFormat.INVALID

    return file_format



@overload
def load_2da(source: str, offset: int = 0):
    ...


@overload
def load_2da(source: bytes, offset: int = 0):
    ...


@overload
def load_2da(source: bytearray, offset: int = 0):
    ...


@overload
def load_2da(source: BinaryReader, offset: int = 0):
    ...


def load_2da(source: SOURCE_TYPES, offset: int = 0) -> TwoDA:
    """
    Returns an TwoDA instance from the source. The file format (binary or xml) is automatically determined before parsing
    the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.

    Raises:
        ValueError: If the file was corrupted or in an unsupported format.

    Returns:
        An TwoDA instance.
    """
    file_format = detect_2da(source, offset)

    try:
        if file_format == FileFormat.BINARY:
            return TwoDABinaryReader(source, offset).load()
        elif file_format == FileFormat.CSV:
            return TwoDACSVReader(source).load()
        else:
            raise ValueError
    except (IOError, ValueError):
        raise ValueError("Tried to load an unsupported or corrupted TwoDA file.")



@overload
def write_2da(twoda: TwoDA, target: str, file_format: FileFormat = FileFormat.BINARY)-> None:
    ...


@overload
def write_2da(twoda: TwoDA, target: bytearray, file_format: FileFormat = FileFormat.BINARY) -> None:
    ...


@overload
def write_2da(twoda: TwoDA, target: BinaryWriter, file_format: FileFormat = FileFormat.BINARY) -> None:
    ...


def write_2da(twoda: TwoDA, target: TARGET_TYPES, file_format: FileFormat = FileFormat.BINARY) -> None:
    """
    Writes the TwoDA data to the target location with the specified format (binary or xml).

    Args:
        twoda: The TwoDA file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        ValueError: If an unsupported FileFormat is passed.
    """
    if file_format == FileFormat.BINARY:
        TwoDABinaryWriter(twoda, target).write()
    elif file_format == FileFormat.CSV:
        TwoDACSVWriter(twoda, target).write()
    else:
        raise ValueError("Unsupported format specified; use BINARY or CSV.")



from __future__ import annotations

import math
import random
from contextlib import suppress
from copy import copy
from typing import Dict, List, Any, Optional, Union, Callable

import glm
from OpenGL.GL import glReadPixels
from OpenGL.raw.GL.VERSION.GL_1_0 import glEnable, GL_TEXTURE_2D, GL_DEPTH_TEST, glClearColor, glClear, \
    GL_COLOR_BUFFER_BIT, GL_DEPTH_BUFFER_BIT, GL_RGBA, GL_BLEND, glBlendFunc, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, \
    glDisable, GL_CULL_FACE, GL_CW, GL_BACK, glCullFace, GL_FRONT_AND_BACK
from OpenGL.raw.GL.VERSION.GL_1_2 import GL_UNSIGNED_INT_8_8_8_8, GL_BGRA
from glm import mat4, vec3, quat, vec4
from pykotor.common.misc import CaseInsensitiveDict
from pykotor.common.module import Module
from pykotor.common.stream import BinaryReader
from pykotor.extract.file import ResourceIdentifier
from pykotor.extract.installation import Installation, SearchLocation
from pykotor.resource.formats.lyt import LYT
from pykotor.resource.formats.tpc import TPC
from pykotor.resource.formats.twoda import read_2da
from pykotor.resource.generics.git import GIT, GITPlaceable, GITCreature, GITDoor, GITTrigger, GITEncounter, \
    GITWaypoint, GITSound, GITStore, GITCamera, GITInstance
from pykotor.resource.type import ResourceType

from pykotor.gl.shader import Shader, KOTOR_VSHADER, KOTOR_FSHADER, Texture, PICKER_FSHADER, PICKER_VSHADER, \
    PLAIN_VSHADER, PLAIN_FSHADER
from pykotor.gl.modelreader import gl_load_mdl, gl_load_stitched_model
from pykotor.gl.model import Model, Cube, Boundary, Empty
from pykotor.gl.models.predefined import STORE_MDL_DATA, STORE_MDX_DATA, WAYPOINT_MDL_DATA, WAYPOINT_MDX_DATA, \
    SOUND_MDL_DATA, SOUND_MDX_DATA, CAMERA_MDL_DATA, CAMERA_MDX_DATA, TRIGGER_MDL_DATA, TRIGGER_MDX_DATA, \
    ENCOUNTER_MDL_DATA, ENCOUNTER_MDX_DATA, ENTRY_MDL_DATA, ENTRY_MDX_DATA, EMPTY_MDL_DATA, EMPTY_MDX_DATA

SEARCH_ORDER_2DA = [SearchLocation.OVERRIDE, SearchLocation.CHITIN]
SEARCH_ORDER = [SearchLocation.CUSTOM_MODULES, SearchLocation.OVERRIDE, SearchLocation.CHITIN]


class Scene:
    SPECIAL_MODELS = ["waypoint", "store", "sound", "camera", "trigger", "encounter"]

    def __init__(self, module: Module, installation: Installation):
        glEnable(GL_TEXTURE_2D)
        glEnable(GL_DEPTH_TEST)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glCullFace(GL_BACK)

        self.installation: Installation = installation
        self.textures: CaseInsensitiveDict[Texture] = CaseInsensitiveDict()
        self.models: CaseInsensitiveDict[Model] = CaseInsensitiveDict()
        self.objects: List[RenderObject] = []
        self.selection: List[RenderObject] = []
        self.module: Module = module
        self.camera: UnfocusedCamera = UnfocusedCamera()

        self.textures["NULL"] = Texture.from_color()

        self.git: Optional[GIT] = None
        self.layout: Optional[LYT] = None
        self.objects: Dict[Any, RenderObject] = {}
        self.clearCacheBuffer: List[ResourceIdentifier] = []

        self.picker_shader: Shader = Shader(PICKER_VSHADER, PICKER_FSHADER)
        self.plain_shader: Shader = Shader(PLAIN_VSHADER, PLAIN_FSHADER)
        self.shader: Shader = Shader(KOTOR_VSHADER, KOTOR_FSHADER)

        self.jumpToEntryLocation()

        self.table_doors = read_2da(installation.resource("genericdoors", ResourceType.TwoDA, SEARCH_ORDER_2DA).data)
        self.table_placeables = read_2da(installation.resource("placeables", ResourceType.TwoDA, SEARCH_ORDER_2DA).data)
        self.table_creatures = read_2da(installation.resource("appearance", ResourceType.TwoDA, SEARCH_ORDER_2DA).data)
        self.table_heads = read_2da(installation.resource("heads", ResourceType.TwoDA, SEARCH_ORDER_2DA).data)

        self.hide_creatures: bool = False
        self.hide_placeables: bool = False
        self.hide_doors: bool = False
        self.hide_triggers: bool = False
        self.hide_encounters: bool = False
        self.hide_waypoints: bool = False
        self.hide_sounds: bool = False
        self.hide_stores: bool = False
        self.hide_cameras: bool = False
        self.hide_sound_boundaries: bool = True
        self.hide_trigger_boundaries: bool = True
        self.hide_encounter_boundaries: bool = True
        self.backface_culling: bool = True

    def buildCache(self, clearCache: bool = False) -> None:
        if clearCache:
            self.objects = {}

        for identifier in self.clearCacheBuffer:
            for creature in copy(self.git.creatures):
                if identifier.resname == creature.resref and identifier.restype == ResourceType.UTC:
                    del self.objects[creature]
            for placeable in copy(self.git.placeables):
                if identifier.resname == placeable.resref and identifier.restype == ResourceType.UTP:
                    del self.objects[placeable]
            for door in copy(self.git.doors):
                if door.resref.get() == identifier.resname and identifier.restype == ResourceType.UTD:
                    del self.objects[door]
            if identifier.restype in [ResourceType.TPC, ResourceType.TGA]:
                del self.textures[identifier.resname]
            if identifier.restype in [ResourceType.MDL, ResourceType.MDX]:
                del self.models[identifier.resname]
            if identifier.restype in [ResourceType.GIT]:
                for instance in self.git.instances():
                    del self.objects[instance]
                self.git = self.module.git().resource()
            if identifier.restype in [ResourceType.LYT]:
                for room in self.layout.rooms:
                    del self.objects[room]
                self.layout = self.module.layout().resource()
        self.clearCacheBuffer = []

        if self.git is None:
            self.git = self.module.git().resource()

        if self.layout is None:
            self.layout = self.module.layout().resource()

        for room in self.layout.rooms:
            if room not in self.objects:
                position = vec3(room.position.x, room.position.y, room.position.z)
                self.objects[room] = RenderObject(room.model, position)

        for door in self.git.doors:
            if door not in self.objects:
                position = vec3(door.position.x, door.position.y, door.position.z)
                rotation = vec3(0, 0, door.bearing)

                try:
                    utd = self.module.door(door.resref.get()).resource()
                    model_name = self.table_doors.get_row(utd.appearance_id).get_string("modelname")
                except Exception:
                    # If failed to load creature models, use an empty model instead
                    model_name = "empty"

                self.objects[door] = RenderObject(model_name, position, rotation, data=door)
            else:
                self.objects[door].set_position(door.position.x, door.position.y, door.position.z)
                self.objects[door].set_rotation(0, 0, door.bearing)

        for placeable in self.git.placeables:
            if placeable not in self.objects:
                position = vec3(placeable.position.x, placeable.position.y, placeable.position.z)
                rotation = vec3(0, 0, placeable.bearing)

                try:
                    utp = self.module.placeable(placeable.resref.get()).resource()
                    model_name = self.table_placeables.get_row(utp.appearance_id).get_string("modelname")
                except Exception:
                    # If failed to load creature models, use an empty model instead
                    model_name = "empty"

                self.objects[placeable] = RenderObject(model_name, position, rotation, data=placeable)
            else:
                self.objects[placeable].set_position(placeable.position.x, placeable.position.y, placeable.position.z)
                self.objects[placeable].set_rotation(0, 0, placeable.bearing)

        for creature in self.git.creatures:
            if creature not in self.objects:
                position = vec3(creature.position.x, creature.position.y, creature.position.z)
                rotation = vec3(0, 0, creature.bearing)

                try:
                    utc = self.module.creature(creature.resref.get()).resource()
                    body_model = self.table_creatures.get_row(utc.appearance_id).get_string("race")
                    obj = RenderObject(body_model, position, rotation, data=creature)

                    head_str = self.table_creatures.get_row(utc.appearance_id).get_string("normalhead")
                    if head_str:
                        head_id = int(head_str)
                        head_model = self.table_heads.get_row(head_id).get_string("head")
                        head_obj = RenderObject(head_model)
                        head_obj.set_transform(self.model(body_model).find("headhook").global_transform())
                        obj.children.append(head_obj)
                except Exception:
                    # If failed to load creature models, use an empty model instead
                    obj = RenderObject("empty", position, rotation, data=creature)

                self.objects[creature] = obj
            else:
                self.objects[creature].set_position(creature.position.x, creature.position.y, creature.position.z)
                self.objects[creature].set_rotation(0, 0, creature.bearing)

        for waypoint in self.git.waypoints:
            if waypoint not in self.objects:
                position = vec3(waypoint.position.x, waypoint.position.y, waypoint.position.z)
                rotation = vec3(0, 0, waypoint.bearing)
                obj = RenderObject("waypoint", position, rotation, data=waypoint)
                self.objects[waypoint] = obj
            else:
                self.objects[waypoint].set_position(waypoint.position.x, waypoint.position.y, waypoint.position.z)
                self.objects[waypoint].set_rotation(0, 0, waypoint.bearing)

        for store in self.git.stores:
            if store not in self.objects:
                position = vec3(store.position.x, store.position.y, store.position.z)
                rotation = vec3(0, 0, store.bearing)
                obj = RenderObject("store", position, rotation, data=store)
                self.objects[store] = obj
            else:
                self.objects[store].set_position(store.position.x, store.position.y, store.position.z)
                self.objects[store].set_rotation(0, 0, store.bearing)

        for sound in self.git.sounds:
            if sound not in self.objects:
                with suppress(Exception):
                    uts = self.module.sound(sound.resref.get()).resource()
                    genBoundary = lambda boundary=Boundary.from_circle(self, uts.max_distance): boundary

                position = vec3(sound.position.x, sound.position.y, sound.position.z)
                rotation = vec3(0, 0, 0)
                obj = RenderObject("sound", position, rotation, data=sound, genBoundary=genBoundary)
                self.objects[sound] = obj
            else:
                self.objects[sound].set_position(sound.position.x, sound.position.y, sound.position.z)
                self.objects[sound].set_rotation(0, 0, 0)

        for encounter in self.git.encounters:
            if encounter not in self.objects:
                position = vec3(encounter.position.x, encounter.position.y, encounter.position.z)
                rotation = vec3(0, 0, 0)
                genBoundary = lambda boundary=Boundary(self, encounter.geometry.points): boundary
                obj = RenderObject("encounter", position, rotation, data=encounter, genBoundary=genBoundary)
                self.objects[encounter] = obj
            else:
                self.objects[encounter].set_position(encounter.position.x, encounter.position.y, encounter.position.z)
                self.objects[encounter].set_rotation(0, 0, 0)

        for trigger in self.git.triggers:
            if trigger not in self.objects:
                position = vec3(trigger.position.x, trigger.position.y, trigger.position.z)
                rotation = vec3(0, 0, 0)
                genBoundary = lambda boundary=Boundary(self, trigger.geometry.points): boundary
                obj = RenderObject("trigger", position, rotation, data=trigger, genBoundary=genBoundary)
                self.objects[trigger] = obj
            else:
                self.objects[trigger].set_position(trigger.position.x, trigger.position.y, trigger.position.z)
                self.objects[trigger].set_rotation(0, 0, 0)

        for camera in self.git.cameras:
            if camera not in self.objects:
                position = vec3(camera.position.x, camera.position.y, camera.position.z)
                rotation = glm.eulerAngles(quat(camera.orientation.x, camera.orientation.y, camera.orientation.z, camera.orientation.w))
                obj = RenderObject("camera", position, rotation, data=camera)
                self.objects[camera] = obj
            else:
                self.objects[camera].set_position(camera.position.x, camera.position.y, camera.position.z)
                euler = glm.eulerAngles(quat(camera.orientation.x, camera.orientation.y, camera.orientation.z, camera.orientation.w))
                self.objects[camera].set_rotation(euler.x, euler.y, euler.z)

        # Detect if GIT still exists; if they do not then remove them from the render list
        for obj in copy(self.objects):
            if isinstance(obj, GITCreature) and obj not in self.git.creatures:
                del self.objects[obj]
            if isinstance(obj, GITPlaceable) and obj not in self.git.placeables:
                del self.objects[obj]
            if isinstance(obj, GITDoor) and obj not in self.git.doors:
                del self.objects[obj]
            if isinstance(obj, GITTrigger) and obj not in self.git.triggers:
                del self.objects[obj]
            if isinstance(obj, GITStore) and obj not in self.git.stores:
                del self.objects[obj]
            if isinstance(obj, GITCamera) and obj not in self.git.cameras:
                del self.objects[obj]
            if isinstance(obj, GITWaypoint) and obj not in self.git.waypoints:
                del self.objects[obj]
            if isinstance(obj, GITEncounter) and obj not in self.git.encounters:
                del self.objects[obj]
            if isinstance(obj, GITSound) and obj not in self.git.sounds:
                del self.objects[obj]

    def render(self) -> None:
        self.buildCache()
        glClearColor(0.5, 0.5, 1, 1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        if self.backface_culling:
            glEnable(GL_CULL_FACE)

        glDisable(GL_BLEND)
        self.shader.use()
        self.shader.set_matrix4("view", self.camera.view())
        self.shader.set_matrix4("projection", self.camera.projection())
        group1 = [obj for obj in self.objects.values() if obj.model not in self.SPECIAL_MODELS]
        for obj in group1:
            self._render_object(self.shader, obj, mat4())

        # Draw all instance types that lack a proper model
        glEnable(GL_BLEND)
        self.plain_shader.use()
        self.plain_shader.set_matrix4("view", self.camera.view())
        self.plain_shader.set_matrix4("projection", self.camera.projection())
        self.plain_shader.set_vector4("color", vec4(0.0, 0.0, 1.0, 0.4))
        group2 = [obj for obj in self.objects.values() if obj.model in self.SPECIAL_MODELS]
        for obj in group2:
            self._render_object(self.plain_shader, obj, mat4())

        # Draw bounding box for selected objects
        self.plain_shader.set_vector4("color", vec4(1.0, 0.0, 0.0, 0.4))
        for obj in self.selection:
            obj.cube(self).draw(self.plain_shader, obj.transform())

        # Draw boundary for selected objects
        glDisable(GL_CULL_FACE)
        self.plain_shader.set_vector4("color", vec4(0.0, 1.0, 0.0, 0.4))
        for obj in self.selection:
            obj.boundary(self).draw(self.plain_shader, obj.transform())

        # Draw non-selected boundaries
        for obj in [obj for obj in self.objects.values() if obj.model == "sound" and not self.hide_sound_boundaries]:
            obj.boundary(self).draw(self.plain_shader, obj.transform())
        for obj in [obj for obj in self.objects.values() if obj.model == "encounter" and not self.hide_encounter_boundaries]:
            obj.boundary(self).draw(self.plain_shader, obj.transform())
        for obj in [obj for obj in self.objects.values() if obj.model == "trigger" and not self.hide_trigger_boundaries]:
            obj.boundary(self).draw(self.plain_shader, obj.transform())

    def _render_object(self, shader: Shader, obj: RenderObject, transform: mat4) -> None:
        if isinstance(obj.data, GITCreature) and self.hide_creatures:
            return
        if isinstance(obj.data, GITPlaceable) and self.hide_placeables:
            return
        if isinstance(obj.data, GITDoor) and self.hide_doors:
            return
        if isinstance(obj.data, GITTrigger) and self.hide_triggers:
            return
        if isinstance(obj.data, GITEncounter) and self.hide_encounters:
            return
        if isinstance(obj.data, GITWaypoint) and self.hide_waypoints:
            return
        if isinstance(obj.data, GITSound) and self.hide_sounds:
            return
        if isinstance(obj.data, GITStore) and self.hide_sounds:
            return
        if isinstance(obj.data, GITCamera) and self.hide_cameras:
            return

        model = self.model(obj.model)
        model.draw(shader, transform * obj.transform())
        for child in obj.children:
            self._render_object(shader, child, obj.transform())

    def picker_render(self) -> None:
        glClearColor(1.0, 1.0, 1.0, 1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        if self.backface_culling:
            glEnable(GL_CULL_FACE)

        self.picker_shader.use()
        self.picker_shader.set_matrix4("view", self.camera.view())
        self.picker_shader.set_matrix4("projection", self.camera.projection())
        instances = list(self.objects.values())
        for i, obj in enumerate(instances):
            int_rgb = instances.index(obj)
            r = int_rgb & 0xFF
            g = (int_rgb >> 8) & 0xFF
            b = (int_rgb >> 16) & 0xFF
            color = vec3(r / 255, g / 255, b / 255)
            self.picker_shader.set_vector3("colorId", color)

            self._picker_render_object(obj, mat4())

    def _picker_render_object(self, obj: RenderObject, transform: mat4) -> None:
        model = self.model(obj.model)
        model.draw(self.picker_shader, transform * obj.transform())
        for child in obj.children:
            self._picker_render_object(child, obj.transform())

    def pick(self, x, y) -> RenderObject:
        self.picker_render()
        pixel = glReadPixels(x, y, 1, 1, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8)[0][0] >> 8
        instances = list(self.objects.values())
        return instances[pixel] if pixel != 0xFFFFFF else None

    def select(self, target: Union[RenderObject, GITInstance], clear_existing: bool = True):
        if clear_existing:
            self.selection.clear()

        if isinstance(target, GITInstance):
            for obj in self.objects.values():
                if obj.data is target:
                    target = obj
                    break

        self.selection.append(target)

    def texture(self, name: str) -> Texture:
        if name not in self.textures:
            try:
                # Check the textures linked to the module first
                tpc = self.module.texture(name).resource() if self.module.texture(name) is not None else None
                # Otherwise just search through all relevant game files
                tpc = self.installation.texture(name, [SearchLocation.OVERRIDE, SearchLocation.TEXTURES_TPA,
                                                       SearchLocation.CHITIN]) if tpc is None else tpc
            except (ValueError, IOError):
                # If an error occurs during the loading process, just use a blank image.
                tpc = TPC()

            self.textures[name] = Texture.from_tpc(tpc) if tpc is not None else Texture.from_color(255, 0, 255)
        return self.textures[name]

    def model(self, name: str) -> Model:
        if name not in self.models:
            if name == "waypoint":
                mdl_data = WAYPOINT_MDL_DATA
                mdx_data = WAYPOINT_MDX_DATA
            elif name == "sound":
                mdl_data = SOUND_MDL_DATA
                mdx_data = SOUND_MDX_DATA
            elif name == "store":
                mdl_data = STORE_MDL_DATA
                mdx_data = STORE_MDX_DATA
            elif name == "entry":
                mdl_data = ENTRY_MDL_DATA
                mdx_data = ENTRY_MDX_DATA
            elif name == "encounter":
                mdl_data = ENCOUNTER_MDL_DATA
                mdx_data = ENCOUNTER_MDX_DATA
            elif name == "trigger":
                mdl_data = TRIGGER_MDL_DATA
                mdx_data = TRIGGER_MDX_DATA
            elif name == "camera":
                mdl_data = CAMERA_MDL_DATA
                mdx_data = CAMERA_MDX_DATA
            elif name == "empty":
                mdl_data = EMPTY_MDL_DATA
                mdx_data = EMPTY_MDX_DATA
            else:
                mdl_search = self.installation.resource(name, ResourceType.MDL, SEARCH_ORDER, capsules=self.module.capsules())
                mdx_search = self.installation.resource(name, ResourceType.MDX, SEARCH_ORDER, capsules=self.module.capsules())
                if mdl_search and mdx_search:
                    mdl_data = mdl_search.data
                    mdx_data = mdx_search.data
                else:
                    mdl_data = EMPTY_MDL_DATA
                    mdx_data = EMPTY_MDX_DATA

            # model = gl_load_mdl(self, BinaryReader.from_bytes(mdl_data, 12), BinaryReader.from_bytes(mdx_data))
            try:
                model = gl_load_stitched_model(self, BinaryReader.from_bytes(mdl_data, 12), BinaryReader.from_bytes(mdx_data))
            except Exception:
                model = gl_load_stitched_model(self, BinaryReader.from_bytes(EMPTY_MDL_DATA, 12), BinaryReader.from_bytes(EMPTY_MDX_DATA))

            self.models[name] = model
        return self.models[name]

    def jumpToEntryLocation(self) -> None:
        point = self.module.info().resource().entry_position
        self.camera.x = point.x
        self.camera.y = point.y
        self.camera.z = point.z + 1.8


class RenderObject:
    def __init__(self, model: str, position: vec3 = None, rotation: vec3 = None, *, data: Any = None,
                 genBoundary: Callable[[], Boundary] = None):
        self.model: str = model
        self.children: List[RenderObject] = []
        self._transform: mat4 = mat4()
        self._position: vec3 = position if position is not None else vec3()
        self._rotation: vec3 = rotation if rotation is not None else vec3()
        self._cube: Optional[Cube] = None
        self._boundary: Optional[Boundary] = None
        self.genBoundary: Optional[Callable[[], Boundary]] = genBoundary
        self.data: Any = data

        self._recalc_transform()

    def transform(self) -> mat4:
        return self._transform

    def set_transform(self, transform: mat4) -> None:
        self._transform = transform
        rotation = quat()
        glm.decompose(transform, vec3(), rotation, self._position, vec3(), vec4())
        self._rotation = glm.eulerAngles(rotation)

    def _recalc_transform(self) -> None:
        self._transform = glm.translate(mat4(), self._position)
        self._transform = self._transform * glm.mat4_cast(quat(self._rotation))

    def position(self) -> vec3:
        return copy(self._position)

    def set_position(self, x: float, y: float, z: float) -> None:
        if self._position.x == x and self._position.y == y and self._position.z == z:
            return

        self._position = vec3(x, y, z)
        self._recalc_transform()

    def rotation(self) -> vec3:
        return copy(self._rotation)

    def set_rotation(self, x: float, y: float, z: float) -> None:
        if self._rotation.x == x and self._rotation.y == y and self._rotation.z == z:
            return

        self._rotation = vec3(x, y, z)
        self._recalc_transform()

    def reset_cube(self) -> None:
        self._cube = None

    def cube(self, scene: Scene) -> Cube:
        if not self._cube:
            min_point = vec3(10000, 10000, 10000)
            max_point = vec3(-10000, -10000, -10000)
            self._cube_rec(scene, mat4(), self, min_point, max_point)
            self._cube = Cube(scene, min_point, max_point)
        return self._cube

    def _cube_rec(self, scene: Scene, transform: mat4, obj: RenderObject, min_point: vec3, max_point: vec3) -> None:
        obj_min, obj_max = scene.model(obj.model).box()
        obj_min = transform * obj_min
        obj_max = transform * obj_max
        min_point.x = min(min_point.x, obj_min.x, obj_max.x)
        min_point.y = min(min_point.y, obj_min.y, obj_max.y)
        min_point.z = min(min_point.z, obj_min.z, obj_max.z)
        max_point.x = max(max_point.x, obj_min.x, obj_max.x)
        max_point.y = max(max_point.y, obj_min.y, obj_max.y)
        max_point.z = max(max_point.z, obj_min.z, obj_max.z)
        for child in obj.children:
            self._cube_rec(scene, transform * child.transform(), child, min_point, max_point)

    def reset_boundary(self) -> None:
        self._boundary = None

    def boundary(self, scene: Scene) -> Boundary | Empty:
        if not self._boundary:
            if self.genBoundary is None:
                self._boundary = Empty(scene)
            else:
                self._boundary = self.genBoundary()
        return self._boundary


class UnfocusedCamera:
    def __init__(self):
        self.x: float = 0.0
        self.y: float = 0.0
        self.z: float = 0.0
        self.pitch: float = math.pi / 2
        self.yaw: float = 0.0
        self.fov: float = 90.0
        self.aspect: float = 16 / 9

    @staticmethod
    def from_focused(camera: FocusedCamera) -> UnfocusedCamera:
        unfocused = UnfocusedCamera()
        unfocused.x = camera.x
        unfocused.y = camera.y
        unfocused.z = camera.z
        unfocused.pitch = camera.pitch
        unfocused.yaw = camera.yaw + math.pi/2
        unfocused.aspect = camera.aspect
        unfocused.fov = camera.fov
        return unfocused

    def view(self) -> mat4:
        up = vec3(0, 0, 1)
        camera = glm.translate(mat4(), vec3(self.x, self.y, self.z))
        camera = glm.rotate(camera, self.yaw, up)
        pitch = glm.vec3(1, 0, 0)
        camera = glm.rotate(camera, self.pitch, pitch)
        view = glm.inverse(camera)
        return view

    def projection(self) -> mat4:
        return glm.perspective(self.fov, self.aspect, 0.1, 5000)

    def translate(self, translation: vec3) -> None:
        self.x += translation.x
        self.y += translation.y
        self.z += translation.z

    def rotate(self, yaw: float, pitch: float):
        self.pitch += pitch
        self.yaw += yaw

        if self.pitch > math.pi:
            self.pitch = math.pi
        elif self.pitch < 0:
            self.pitch = 0

    def forward(self, ignoreZ: bool = True) -> vec3:
        eye_x = -math.cos(self.yaw) * math.cos(self.pitch - math.pi/2)
        eye_y = math.sin(self.yaw) * math.cos(self.pitch - math.pi/2)
        eye_z = 0 if ignoreZ else math.sin(self.pitch - math.pi/2)
        return -glm.normalize(vec3(eye_y, eye_x, eye_z))

    def sideward(self, ignoreZ: bool = True) -> vec3:
        return glm.normalize(glm.cross(self.forward(ignoreZ), vec3(0.0, 0.0, 1.0)))

    def upward(self, ignoreXY: bool = True) -> vec3:
        if not ignoreXY:
            forward = self.forward(False)
            sideward = self.sideward(False)
            cross = glm.cross(forward, sideward)
            return glm.normalize(cross)
        else:
            return glm.normalize(vec3(0, 0, 1))


class FocusedCamera:
    def __init__(self):
        self.x: float = 40.0
        self.y: float = 130.0
        self.z: float = 0.5
        self.pitch: float = math.pi / 2
        self.yaw: float = 0.0
        self.distance: float = 10.0
        self.fov: float = 90.0
        self.aspect: float = 16 / 9

    @staticmethod
    def from_unfocused(camera: UnfocusedCamera) -> FocusedCamera:
        focused = FocusedCamera()
        focused.x = camera.x
        focused.y = camera.y
        focused.z = camera.z
        focused.pitch = camera.pitch
        focused.yaw = camera.yaw - math.pi/2
        focused.aspect = camera.aspect
        focused.fov = camera.fov
        return focused

    def view(self) -> mat4:
        up = vec3(0, 0, 1)
        pitch = glm.vec3(1, 0, 0)

        x, y, z = self.x, self.y, self.z
        x += math.cos(self.yaw) * math.cos(self.pitch - math.pi/2) * self.distance
        y += math.sin(self.yaw) * math.cos(self.pitch - math.pi/2) * self.distance
        z += math.sin(self.pitch - math.pi/2) * self.distance

        camera = glm.translate(mat4(), vec3(x, y, z))
        camera = glm.rotate(camera, self.yaw + math.pi/2, up)
        camera = glm.rotate(camera, math.pi - self.pitch, pitch)
        view = glm.inverse(camera)
        return view

    def projection(self) -> mat4:
        return glm.perspective(self.fov, self.aspect, 0.1, 5000)

    def translate(self, translation: vec3) -> None:
        self.x += translation.x
        self.y += translation.y
        self.z += translation.z

    def rotate(self, yaw: float, pitch: float):
        self.pitch += pitch
        self.yaw += yaw

        if self.pitch > math.pi - 0.001:
            self.pitch = math.pi - 0.001
        elif self.pitch < 0.001:
            self.pitch = 0.001

    def forward(self, ignoreZ: bool = True) -> vec3:
        eye_x = math.cos(self.yaw) * math.cos(self.pitch - math.pi / 2)
        eye_y = math.sin(self.yaw) * math.cos(self.pitch - math.pi / 2)
        eye_z = 0 if ignoreZ else math.sin(self.pitch - math.pi / 2)
        return glm.normalize(-vec3(eye_x, eye_y, eye_z))

    def sideward(self, ignoreZ: bool = True) -> vec3:
        return glm.normalize(glm.cross(self.forward(ignoreZ), vec3(0.0, 0.0, 1.0)))

    def upward(self, ignoreXY: bool = True) -> vec3:
        if not ignoreXY:
            forward = self.forward(False)
            sideward = self.sideward(False)
            cross = glm.cross(forward, sideward)
            return glm.normalize(cross)
        else:
            return glm.normalize(vec3(0, 0, 1))



import os
from unittest import TestCase

from pykotor.resource.type import ResourceType

from pykotor.common.misc import filepath_info
from pykotor.extract.installation import Installation


class TestMisc(TestCase):
    def test_filepath_info(self):
        self.assertEqual(("lightsaber01", ResourceType.UTI), filepath_info("C:/somefolder/lightsaber01.uti"))
        self.assertEqual(("lightsaber01", ResourceType.UTI_XML), filepath_info("C:/somefolder/lightsaber01.uti.xml"))



from typing import overload, Union

from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.lip import LIP, LIPBinaryReader, LIPXMLReader, LIPBinaryWriter
from pykotor.resource.formats.lip.io_xml import LIPXMLWriter
from pykotor.resource.type import FileFormat, SOURCE_TYPES, TARGET_TYPES


@overload
def detect_lip(source: str, offset: int = 0):
    ...


@overload
def detect_lip(source: bytes, offset: int = 0):
    ...


@overload
def detect_lip(source: bytearray, offset: int = 0):
    ...


@overload
def detect_lip(source: BinaryReader, offset: int = 0):
    ...


def detect_lip(source: SOURCE_TYPES, offset: int = 0) -> FileFormat:
    """
    Returns what format the LIP data is believed to be in. This function performs a basic check and does not guarantee
    accuracy of the result or integrity of the data.

    Args:
        source: Source of the LIP data.
        offset: Offset into the data.

    Returns:
        The format of the LIP data.
    """
    try:
        if isinstance(source, str):
            with BinaryReader.from_file(source, offset) as reader:
                file_format = FileFormat.BINARY if reader.read_string(4) == "LIP " else FileFormat.XML
        elif isinstance(source, bytes) or isinstance(source, bytearray):
            file_format = FileFormat.BINARY if source[:4].decode('ascii', 'ignore') == "LIP " else FileFormat.XML
        elif isinstance(source, BinaryReader):
            file_format = FileFormat.BINARY if source.read_string(4) == "LIP " else FileFormat.XML
            source.skip(-4)
        else:
            file_format = FileFormat.INVALID
    except IOError:
        file_format = FileFormat.INVALID

    return file_format


def load_lip(source: SOURCE_TYPES, offset: int = 0) -> LIP:
    """
    Returns an LIP instance from the source. The file format (binary or xml) is automatically determined before parsing
    the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.

    Raises:
        ValueError: If the file was corrupted or in an unsupported format.

    Returns:
        An LIP instance.
    """
    file_format = detect_lip(source, offset)

    try:
        if file_format == FileFormat.BINARY:
            return LIPBinaryReader(source, offset).load()
        elif file_format == FileFormat.XML:
            return LIPXMLReader(source).load()
        else:
            raise ValueError
    except (IOError, ValueError):
        raise ValueError("Tried to load an unsupported or corrupted LIP file.")


def write_lip(lip: LIP, target: TARGET_TYPES, file_format: FileFormat = FileFormat.BINARY) -> None:
    """
    Writes the LIP data to the target location with the specified format (binary or xml).

    Args:
        lip: The LIP file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        ValueError: If an unsupported FileFormat is passed.
    """
    if file_format == FileFormat.BINARY:
        LIPBinaryWriter(lip, target).write()
    elif file_format == FileFormat.XML:
        LIPXMLWriter(lip, target).write()
    else:
        raise ValueError("Unsupported format specified; use BINARY or XML.")



from unittest import TestCase

from pykotor.resource.formats.vis import VISAsciiReader, VIS
from pykotor.resource.formats.vis.auto import write_vis, load_vis
from pykotor.resource.type import FileFormat

ASCII_TEST_FILE = "../../files/test.vis"


class TestLIP(TestCase):
    def test_binary_io(self):
        vis = VISAsciiReader(ASCII_TEST_FILE).load()
        self.validate_io(vis)

        data = bytearray()
        write_vis(vis, data, FileFormat.ASCII)
        vis = load_vis(data)
        self.validate_io(vis)

    def validate_io(self, vis: VIS):
        self.assertTrue(vis.get_visible("room_01", "room_02"))
        self.assertTrue(vis.get_visible("room_01", "room_03"))
        self.assertTrue(vis.get_visible("room_01", "room_04"))

        self.assertTrue(vis.get_visible("room_02", "room_01"))
        self.assertFalse(vis.get_visible("room_02", "room_03"))
        self.assertFalse(vis.get_visible("room_02", "room_04"))

        self.assertTrue(vis.get_visible("room_03", "room_01"))
        self.assertTrue(vis.get_visible("room_03", "room_04"))
        self.assertFalse(vis.get_visible("room_03", "room_02"))

        self.assertTrue(vis.get_visible("room_04", "room_01"))
        self.assertTrue(vis.get_visible("room_04", "room_03"))
        self.assertFalse(vis.get_visible("room_04", "room_02"))



from __future__ import annotations

import csv
import io
from typing import Optional

from pykotor.resource.formats.twoda.data import TwoDA
from pykotor.resource.type import TARGET_TYPES, SOURCE_TYPES, ResourceReader, ResourceWriter


class TwoDACSVReader(ResourceReader):
    def __init__(self, source: SOURCE_TYPES, offset: int = 0, size: int = 0):
        super().__init__(source, offset, size)
        self._csv: csv.reader = csv.reader(io.StringIO(self._reader.read_bytes(self._size).decode()))
        self._twoda: Optional[TwoDA] = None

    def load(self, auto_close: bool = True) -> TwoDA:
        self._twoda = TwoDA()

        headers = next(self._csv)[1:]
        for header in headers:
            self._twoda.add_column(header)

        for row in self._csv:
            label = int(row[:1][0])
            cells = dict(zip(headers, row[1:]))
            self._twoda.add_row(label, cells)

        if auto_close:
            self._reader.close()

        return self._twoda


class TwoDACSVWriter(ResourceWriter):
    def __init__(self, twoda: TwoDA, target: TARGET_TYPES):
        super().__init__(target)
        self._twoda: TwoDA = twoda
        self._csv_string = io.StringIO("")
        self._csv_writer = csv.writer(self._csv_string)

    def write(self, auto_close: bool = True) -> None:
        headers = self._twoda.get_headers()

        insert = [""]
        for header in headers:
            insert.append(header)
        self._csv_writer.writerow(insert)

        for row in self._twoda:
            insert = [str(row.label())]
            for header in headers:
                insert.append(row.get_string(header))
            self._csv_writer.writerow(insert)

        data = self._csv_string.getvalue().encode('ascii')
        self._writer.write_bytes(data)

        if auto_close:
            self._writer.close()



from unittest import TestCase

from pykotor.resource.formats.rim import RIM, RIMBinaryReader, write_rim, RIMBinaryWriter, load_rim
from pykotor.resource.type import ResourceType

BINARY_TEST_FILE = "../../files/test.rim"


class TestRIM(TestCase):
    def test_binary_io(self):
        rim = RIMBinaryReader(BINARY_TEST_FILE).load()
        self.validate_io(rim)

        data = bytearray()
        write_rim(rim, data)
        rim = load_rim(data)
        self.validate_io(rim)

    def validate_io(self, rim: RIM):
        self.assertEqual(len(rim), 3)
        self.assertEqual(rim.get("1", ResourceType.TXT), b'abc')
        self.assertEqual(rim.get("2", ResourceType.TXT), b'def')
        self.assertEqual(rim.get("3", ResourceType.TXT), b'ghi')



import os
from unittest import TestCase

from pykotor.common.module import Module
from pykotor.resource.type import ResourceType

from pykotor.common.misc import filepath_info
from pykotor.extract.installation import Installation


class TestModule(TestCase):
    def test_module(self):
        installation = Installation("C:/Program Files (x86)/Steam/steamapps/common/swkotor/")
        module = Module("danm13", installation)
        module.layout.resource()



'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GL import _types as _cs
# End users want this...
from OpenGL.raw.GL._types import *
from OpenGL.raw.GL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GL_HP_texture_lighting'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GL,'GL_HP_texture_lighting',error_checker=_errors._error_checker)
GL_TEXTURE_LIGHTING_MODE_HP=_C('GL_TEXTURE_LIGHTING_MODE_HP',0x8167)
GL_TEXTURE_POST_SPECULAR_HP=_C('GL_TEXTURE_POST_SPECULAR_HP',0x8168)
GL_TEXTURE_PRE_SPECULAR_HP=_C('GL_TEXTURE_PRE_SPECULAR_HP',0x8169)




from typing import Union

from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.lyt import LYT, LYTAsciiWriter, LYTAsciiReader
from pykotor.resource.type import FileFormat


def load_lyt(source: Union[str, bytes, bytearray, BinaryReader], offset: int = 0) -> LYT:
    """
    Returns an LYT instance from the source. The file format (binary or xml) is automatically determined before parsing
    the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.

    Raises:
        ValueError: If the file was corrupted or in an unsupported format.

    Returns:
        An LYT instance.
    """
    try:
        return LYTAsciiReader(source, offset).load()
    except IOError:
        raise ValueError("Tried to load an unsupported or corrupted LYT file.")


def write_lyt(lyt: LYT, target: Union[str, bytearray, BinaryReader], file_format: FileFormat = FileFormat.ASCII) -> None:
    """
    Writes the LYT data to the target location with the specified format (ascii only).

    Args:
        lyt: The LYT file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        ValueError: If an unsupported FileFormat is passed.
    """
    if file_format == FileFormat.ASCII:
        LYTAsciiWriter(lyt, target).write()
    else:
        raise ValueError("Unsupported format specified; use BINARY or XML.")



from unittest import TestCase

from pykotor.resource.formats.twoda import TwoDA, write_2da, TwoDABinaryReader, load_2da, detect_2da, TwoDACSVReader
from pykotor.resource.type import ResourceType, FileFormat

BINARY_TEST_FILE = "../../files/test.2da"
CSV_TEST_FILE = "../../files/test.2da.csv"


class TestTwoDA(TestCase):
    def test_binary_io(self):
        self.assertEqual(detect_2da(BINARY_TEST_FILE), FileFormat.BINARY)
        twoda = TwoDABinaryReader(BINARY_TEST_FILE).load()
        self.validate_io(twoda)

        data = bytearray()
        write_2da(twoda, data, FileFormat.BINARY)
        twoda = load_2da(data)
        self.validate_io(twoda)

    def test_csv_io(self):
        self.assertEqual(detect_2da(CSV_TEST_FILE), FileFormat.CSV)

        twoda = TwoDACSVReader(CSV_TEST_FILE).load()
        self.validate_io(twoda)

        data = bytearray()
        write_2da(twoda, data, FileFormat.CSV)
        twoda = load_2da(data)
        self.validate_io(twoda)

    def validate_io(self, twoda):
        self.assertEqual("abc", twoda.get_cell(0, "col1"))
        self.assertEqual("def", twoda.get_cell(0, "col2"))
        self.assertEqual("ghi", twoda.get_cell(0, "col3"))

        self.assertEqual("def", twoda.get_cell(1, "col1"))
        self.assertEqual("ghi", twoda.get_cell(1, "col2"))
        self.assertEqual("123", twoda.get_cell(1, "col3"))

        self.assertEqual("123", twoda.get_cell(2, "col1"))
        self.assertEqual("", twoda.get_cell(2, "col2"))
        self.assertEqual("abc", twoda.get_cell(2, "col3"))



from pykotor.resource.formats.ltr.data import LTR
from pykotor.resource.formats.ltr.io_binary import LTRBinaryReader, LTRBinaryWriter
from pykotor.resource.type import SOURCE_TYPES, FileFormat, TARGET_TYPES


def load_ltr(source: SOURCE_TYPES, offset: int = 0) -> LTR:
    """
    Returns an LTR instance from the source.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.

    Raises:
        ValueError: If the file was corrupted or in an unsupported format.

    Returns:
        An LTR instance.
    """
    try:
        return LTRBinaryReader(source, offset).load()
    except (IOError, ValueError):
        raise ValueError("Tried to load an unsupported or corrupted LTR file.")


def write_ltr(ltr: LTR, target: TARGET_TYPES, file_format: FileFormat = FileFormat.BINARY) -> None:
    """
    Writes the LTR data to the target location with the specified format (binary only).

    Args:
        ltr: The LTR file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        ValueError: If an unsupported FileFormat is passed.
    """
    if file_format == FileFormat.BINARY:
        LTRBinaryWriter(ltr, target).write()
    else:
        raise ValueError("Unsupported format specified; use BINARY or XML.")



import os
from unittest import TestCase

from pykotor.common.stream import BinaryReader

from pykotor.extract.capsule import Capsule
from pykotor.resource.formats.tpc import load_tpc, write_tpc

from pykotor.resource.type import ResourceType, FileFormat

from pykotor.extract.installation import Installation
from pykotor.tools import model


class TestInstallation(TestCase):
    def setUp(self) -> None:
        self.path = os.environ.get('K1_PATH')
        self.installation = Installation(self.path)

    def test_other(self):
        path = "C:/Users/hugin/Desktop/ext/clone/hoth01_01a.mdl"
        data = BinaryReader.load_file(path)
        data2 = model.rename(data, "WOLOLOOLOLOLOLOLOOLOLOLO")
        print(data2)

    def test_resource(self):
        # testing capsules param
        path, res = self.installation.resource("m13aa", ResourceType.ARE, capsules=[Capsule(self.path + "modules/danm13.rim")], skip_modules=True)
        self.assertIsNotNone(res)

        # testing installation modules
        path, res = self.installation.resource("m01aa", ResourceType.ARE, skip_chitin=True, skip_override=True)
        self.assertIsNotNone(res)
        path, res = self.installation.resource("m13aa", ResourceType.ARE, skip_modules=True)
        self.assertIsNone(res)

        # testing folders param
        path, res = self.installation.resource("biowarelogo", ResourceType.TGA, folders=["C:/Users/hugin/Desktop/ext"])
        self.assertIsNotNone(res)
        path, res = self.installation.resource("biowarelogo", ResourceType.TGA)
        self.assertIsNone(res)

        # testing chitin
        path, res = self.installation.resource("nwscript", ResourceType.NSS, skip_override=True, skip_modules=True)
        self.assertIsNotNone(res)

        # testing installation override
        path, res = self.installation.resource("nwscript", ResourceType.NSS, skip_chitin=True, skip_modules=True)

    def test_locate(self):
        print(self.installation.locate("nwscript", ResourceType.NSS))
        print(self.installation.locate("m13aa", ResourceType.ARE))
        print(self.installation.locate("c_bantha", ResourceType.UTC, folders=["C:/Users/hugin/Desktop/ext"]))
        print(self.installation.locate("C_Gammorean01", ResourceType.TPC))
        #print(self.installation.locate("new", ResourceType.GFF, capsules=[Capsule("C:/Users/hugin/Desktop/ext/blank.erf")]))




from __future__ import annotations

import io
from typing import Optional
from xml.etree import ElementTree

from pykotor.resource.formats.lip import LIP, LIPShape
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceReader, ResourceWriter


class LIPXMLReader(ResourceReader):
    def __init__(self, source: SOURCE_TYPES, offset: int = 0, size: int = 0):
        super().__init__(source, offset, size)
        self._xml_root: ElementTree.Element = ElementTree.parse(io.StringIO(self._reader.read_bytes(self._size).decode())).getroot()
        self._lip: Optional[LIP] = None

    def load(self, auto_close: bool = True) -> LIP:
        if self._xml_root.tag != "lip":
            raise TypeError("The XML file that was loaded was not a valid LIP.")

        self._lip = LIP()

        self._lip.length = float(self._xml_root.get("duration"))

        for subelement in self._xml_root:
            time = float(subelement.get("time"))
            shape = LIPShape(int(subelement.get("shape")))
            self._lip.add(time, shape)

        if auto_close:
            self._reader.close()

        return self._lip


class LIPXMLWriter(ResourceWriter):
    def __init__(self, lip: LIP, target: TARGET_TYPES):
        super().__init__(target)
        self._lip = lip
        self._xml_root: ElementTree.Element = ElementTree.Element("lip")

    def write(self, auto_close: bool = True) -> None:
        self._xml_root.set("duration", str(self._lip.length))

        for keyframe in self._lip:
            ElementTree.SubElement(self._xml_root,
                                   "keyframe",
                                   time=str(keyframe.time),
                                   shape=str(keyframe.shape.value))

        ElementTree.indent(self._xml_root)
        self._writer.write_bytes(ElementTree.tostring(self._xml_root))

        if auto_close:
            self._writer.close()



from typing import Union

from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.vis import VIS, VISAsciiWriter, VISAsciiReader
from pykotor.resource.type import FileFormat


def load_vis(source: Union[str, bytes, bytearray, BinaryReader], offset: int = 0) -> VIS:
    """
    Returns an VIS instance from the source. The file format (ascii only) is automatically determined before parsing
    the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.

    Raises:
        ValueError: If the file was corrupted or in an unsupported format.

    Returns:
        An VIS instance.
    """
    try:
        return VISAsciiReader(source, offset).load()
    except IOError:
        raise ValueError("Tried to load an unsupported or corrupted VIS file.")


def write_vis(vis: VIS, target: Union[str, bytearray, BinaryReader], file_format: FileFormat = FileFormat.ASCII) -> None:
    """
    Writes the VIS data to the target location with the specified format (ascii only).

    Args:
        vis: The VIS file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        ValueError: If an unsupported FileFormat is passed.
    """
    if file_format == FileFormat.ASCII:
        VISAsciiWriter(vis, target).write()
    else:
        raise ValueError("Unsupported format specified; use ASCII.")



'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GLES2 import _types as _cs
# End users want this...
from OpenGL.raw.GLES2._types import *
from OpenGL.raw.GLES2 import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GLES2_EXT_base_instance'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GLES2,'GLES2_EXT_base_instance',error_checker=_errors._error_checker)

@_f
@_p.types(None,_cs.GLenum,_cs.GLint,_cs.GLsizei,_cs.GLsizei,_cs.GLuint)
def glDrawArraysInstancedBaseInstanceEXT(mode,first,count,instancecount,baseinstance):pass
@_f
@_p.types(None,_cs.GLenum,_cs.GLsizei,_cs.GLenum,ctypes.c_void_p,_cs.GLsizei,_cs.GLuint)
def glDrawElementsInstancedBaseInstanceEXT(mode,count,type,indices,instancecount,baseinstance):pass
@_f
@_p.types(None,_cs.GLenum,_cs.GLsizei,_cs.GLenum,ctypes.c_void_p,_cs.GLsizei,_cs.GLint,_cs.GLuint)
def glDrawElementsInstancedBaseVertexBaseInstanceEXT(mode,count,type,indices,instancecount,basevertex,baseinstance):pass



from unittest import TestCase

from pykotor.common.language import Language
from pykotor.common.misc import ResRef
from pykotor.resource.formats.tlk import TLK, TLKEntry, detect_tlk, TLKBinaryReader, write_tlk, load_tlk
from pykotor.resource.type import FileFormat

BINARY_TEST_FILE = "../../files/test.tlk"
XML_TEST_FILE = "../../files/test.tlk.xml"


class TestTLK(TestCase):
    def test_binary_io(self):
        self.assertEqual(detect_tlk(BINARY_TEST_FILE), FileFormat.BINARY)

        tlk = TLKBinaryReader(BINARY_TEST_FILE).load()
        self.validate_io(tlk)

        data = bytearray()
        write_tlk(tlk, data, FileFormat.BINARY)
        tlk = load_tlk(data)
        self.validate_io(tlk)

    def validate_io(self, tlk: TLK):
        self.assertIs(tlk.language, Language.ENGLISH)

        self.assertEqual(TLKEntry("abcdef", ResRef("resref01")), tlk[0])
        self.assertEqual(TLKEntry("ghijklmnop", ResRef("resref02")), tlk[1])
        self.assertEqual(TLKEntry("qrstuvwxyz", ResRef("")), tlk[2])

    def test_resize(self):
        tlk = TLKBinaryReader(BINARY_TEST_FILE).load()
        self.assertEqual(len(tlk), 3)
        tlk.resize(4)
        self.assertEqual(len(tlk), 4)
        self.assertEqual(TLKEntry("qrstuvwxyz", ResRef("")), tlk[2])
        self.assertEqual(TLKEntry("", ResRef("")), tlk[3])
        tlk.resize(1)
        self.assertEqual(len(tlk), 1)
        self.assertEqual(TLKEntry("abcdef", ResRef("resref01")), tlk.get(0))
        self.assertIsNone(tlk.get(1))



from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.tlk import TLK, TLKBinaryReader, TLKXMLReader, TLKBinaryWriter
from pykotor.resource.formats.tlk.io_tlk_json import TLKJSONReader, TLKJSONWriter
from pykotor.resource.formats.tlk.io_tlk_xml import TLKXMLWriter
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceType


def detect_tlk(
        source: SOURCE_TYPES,
        offset: int = 0
) -> ResourceType:
    """
    Returns what format the TLK data is believed to be in. This function performs a basic check and does not guarantee
    accuracy of the result or integrity of the data.

    Args:
        source: Source of the TLK data.
        offset: Offset into the data.

    Raises:
        IOError: If an error occured reading the file.

    Returns:
        The format of the TLK data.
    """

    def check(
            first4
    ):
        if first4 == "TLK ":
            return ResourceType.TLK
        elif "{" in first4:
            return ResourceType.TLK_JSON
        elif "<" in first4:
            return ResourceType.TLK_XML
        else:
            return ResourceType.INVALID

    if isinstance(source, str):
        with BinaryReader.from_file(source, offset) as reader:
            file_format = check(reader.read_string(4))
    elif isinstance(source, bytes) or isinstance(source, bytearray):
        file_format = check(source[:4].decode('ascii', 'ignore'))
    elif isinstance(source, BinaryReader):
        file_format = check(source.read_string(4))
        source.skip(-4)
    else:
        file_format = ResourceType.INVALID

    return file_format


def read_tlk(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> TLK:
    """
    Returns an TLK instance from the source. The file format (TLK, TLK_XML or TLK_JSON) is automatically determined
    before parsing the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.
        size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

    Raises:
        IOError: If an error occured reading the file.
        ValueError: If unable to determine the file format or the file data was corrupted.

    Returns:
        An TLK instance.
    """
    file_format = detect_tlk(source, offset)

    if file_format == ResourceType.TLK:
        return TLKBinaryReader(source, offset, size).load()
    elif file_format == ResourceType.TLK_XML:
        return TLKXMLReader(source, offset, size).load()
    elif file_format == ResourceType.TLK_JSON:
        return TLKJSONReader(source, offset, size).load()
    else:
        raise ValueError("Unable to determine the file format.")


def write_tlk(
        tlk: TLK,
        target: TARGET_TYPES,
        file_format: ResourceType = ResourceType.TLK
) -> None:
    """
    Writes the TLK data to the target location with the specified format (TLK, TLK_XML or TLK_JSON).

    Args:
        tlk: The TLK file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        IOError: If an error occured writing to the file.
        ValueError: If an unsupported file format was given.
    """
    if file_format == ResourceType.TLK:
        TLKBinaryWriter(tlk, target).write()
    elif file_format == ResourceType.TLK_XML:
        TLKXMLWriter(tlk, target).write()
    elif file_format == ResourceType.TLK_JSON:
        TLKJSONWriter(tlk, target).write()
    else:
        raise ValueError("Unsupported format specified; use TLK or TLK_XML.")


def bytes_tlk(
        tlk: TLK,
        file_format: ResourceType = ResourceType.TLK
) -> bytes:
    """
    Returns the TLK data in the specified format (TLK or TLK_XML or TLK_JSON) as a bytes object.

    This is a convience method that wraps the write_tlk() method.

    Args:
        tlk: The target TLK object.
        file_format: The file format.

    Raises:
        ValueError: If an unsupported file format was given.

    Returns:
        The TLK data.
    """
    data = bytearray()
    write_tlk(tlk, data, file_format)
    return data



"""
This module handles classes relating to editing TLK files.
"""
from __future__ import annotations

from typing import List

from pykotor.resource.type import ResourceType

from pykotor.common.language import Language
from pykotor.common.misc import ResRef


class TLK:

    BINARY_TYPE = ResourceType.TLK

    def __init__(self):
        self.entries: List[TLKEntry] = []
        self.language: Language = Language.ENGLISH

    def __len__(self):
        """
        Returns the number of stored entries.
        """
        return len(self.entries)

    def __iter__(self):
        """
        Iterates through the stored entry with each iteration yielding a stringref and the corresponding entry data.
        """
        for stringref, entry in enumerate(self.entries):
            yield stringref, entry

    def __getitem__(self, item):
        """
        Returns an entry for the specified stringref.

        Args:
            item: The stringref.

        Raises:
            IndexError: If the stringref does not exist.

        Returns:
            The corresponding TLKEntry.
        """
        if not isinstance(item, int):
            return NotImplemented
        return self.entries[item]

    def get(self, stringref: int) -> TLKEntry:
        """
        Returns an entry for the specified stringref if it exists, otherwise returns None.

        Args:
            stringref: The stringref.

        Returns:
            The corresponding TLKEntry or None.
        """
        return self.entries[stringref] if 0 <= stringref < len(self) else None

    def resize(self, size: int) -> None:
        """
        Resizes the number of entries to the specified size.

        Args:
            size: The new number of entries.
        """
        if len(self) > size:
            self.entries = self.entries[:size]
        else:
            self.entries.extend([TLKEntry("", ResRef.from_blank()) for _ in range(len(self), size)])


class TLKEntry:
    def __init__(self, text: str, voiceover: ResRef):
        self.text: str = text
        self.voiceover: ResRef = voiceover

    def __eq__(self, other):
        """
        Returns True if the text and voiceover match.
        """
        if not isinstance(other, TLKEntry):
            return NotImplemented
        return other.text == self.text and other.voiceover == self.voiceover



from unittest import TestCase

from pykotor.common.geometry import Vector3, Vector4
from pykotor.resource.formats.lyt import LYTAsciiReader, LYTRoom, LYTTrack, LYT, LYTObstacle, LYTDoorHook
from pykotor.resource.formats.lyt.auto import write_lyt, load_lyt
from pykotor.resource.type import FileFormat

ASCII_TEST_FILE = "../../files/test.lyt"


class TestLYT(TestCase):
    def test_binary_io(self):
        lyt = LYTAsciiReader(ASCII_TEST_FILE).load()
        self.validate_io(lyt)

        data = bytearray()
        write_lyt(lyt, data, FileFormat.ASCII)
        lyt = load_lyt(data)
        self.validate_io(lyt)

    def validate_io(self, lyt: LYT):
        self.assertEqual(lyt.rooms[0], LYTRoom("M17mg_01a", Vector3(100.0, 100.0, 0.0)))
        self.assertEqual(lyt.rooms[1], LYTRoom("M17mg_01b", Vector3(100.0, 100.0, 0.0)))
        self.assertEqual(lyt.tracks[0], LYTTrack("M17mg_MGT01", Vector3(0.0, 0.0, 0.0)))
        self.assertEqual(lyt.tracks[1], LYTTrack("M17mg_MGT02", Vector3(112.047, 209.04, 0.0)))
        self.assertEqual(lyt.obstacles[0], LYTObstacle("M17mg_MGO01", Vector3(103.309, 3691.61, 0.0)))
        self.assertEqual(lyt.obstacles[1], LYTObstacle("M17mg_MGO02", Vector3(118.969, 3688.0, 0.0)))
        self.assertEqual(lyt.doorhooks[0], LYTDoorHook("M02ac_02h", "door_01", Vector3(170.475, 66.375, 0.0),
                                                       Vector4(0.707107, 0.0, 0.0, -0.707107)))
        self.assertEqual(lyt.doorhooks[1],
                         LYTDoorHook("M02ac_02a", "door_06", Vector3(90.0, 129.525, 0.0), Vector4(1.0, 0.0, 0.0, 0.0)))



from typing import overload, Union

from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.tpc import TPC, TPCBinaryReader, TPCBinaryWriter, TPCTGAWriter, TPCBMPWriter, TPCTGAReader
from pykotor.resource.type import FileFormat, SOURCE_TYPES, TARGET_TYPES


def detect_tpc(source: SOURCE_TYPES, offset: int = 0) -> FileFormat:
    """
    Returns what format the TPC data is believed to be in. This function performs a basic check and does not guarantee
    accuracy of the result or integrity of the data.

    Args:
        source: Source of the TPC data.
        offset: Offset into the source data.

    Returns:
        The format of the TPC data.
    """
    def do_check(first100: bytes) -> FileFormat:
        file_format = FileFormat.BINARY
        for i in range(15, 100):
            if first100[i] != 0:
                file_format = FileFormat.TGA
        return file_format

    try:
        if isinstance(source, str):
            with BinaryReader.from_file(source, offset) as reader:
                file_format = do_check(reader.read_bytes(100))
        elif isinstance(source, bytes) or isinstance(source, bytearray):
            file_format = do_check(source[:100])
        elif isinstance(source, BinaryReader):
            file_format = do_check(source.read_bytes(100))
            source.skip(-100)
        else:
            file_format = FileFormat.INVALID
    except IOError:
        file_format = FileFormat.INVALID

    return file_format


def load_tpc(source: SOURCE_TYPES, offset: int = 0) -> TPC:
    """
    Returns an TPC instance from the source. The file format (tpc or tga) is automatically determined before
    parsing the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.

    Raises:
        ValueError: If the file was corrupted or in an unsupported format.

    Returns:
        An TPC instance.
    """
    file_format = detect_tpc(source, offset)

    try:
        if file_format == FileFormat.BINARY:
            return TPCBinaryReader(source, offset).load()
        elif file_format == FileFormat.TGA:
            return TPCTGAReader(source).load()
        else:
            raise ValueError
    except (IOError, ValueError):
        raise ValueError("Tried to load an unsupported or corrupted TPC file.")


def write_tpc(tpc: TPC, target: TARGET_TYPES, file_format: FileFormat = FileFormat.BINARY) -> None:
    """
    Writes the TPC data to the target location with the specified format (tpc, tga, bmp).

    Args:
        tpc: The TPC file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        ValueError: If an unsupported FileFormat is passed.
    """
    if file_format == FileFormat.TGA:
        TPCTGAWriter(tpc, target).write()
    elif file_format == FileFormat.BMP:
        TPCBMPWriter(tpc, target).write()
    elif file_format == FileFormat.BINARY:
        TPCBinaryWriter(tpc, target).write()
    else:
        raise ValueError("Unsupported format specified; use BINARY or XML.")




from __future__ import annotations

import io
from typing import Optional
from xml.etree import ElementTree

from pykotor.resource.formats.tlk.data import TLK
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceReader, ResourceWriter


class TLKXMLReader(ResourceReader):
    def __init__(self, source: SOURCE_TYPES, offset: int = 0, size: int = 0):
        super().__init__(source, offset, size)
        self._xml_root: ElementTree.Element = ElementTree.parse(io.StringIO(self._reader.read_bytes(self._size).decode())).getroot()
        self._tlk: Optional[TLK] = None

    def load(self, auto_close: bool = True) -> TLK:
        self._tlk = TLK()

        # TODO

        if auto_close:
            self._reader.close()

        return self._tlk


class TLKXMLWriter(ResourceWriter):
    def __init__(self, tlk: TLK, target: TARGET_TYPES):
        super().__init__(target)
        self.xml_root: ElementTree.Element = ElementTree.Element("xml")
        self.tlk: TLK = tlk

    def write(self, auto_close: bool = True) -> None:
        # TODO

        ElementTree.indent(self.xml_root)
        self._writer.write_bytes(ElementTree.tostring(self.xml_root))

        if auto_close:
            self._writer.close()



from __future__ import annotations

import os.path
from contextlib import suppress
from typing import List, TypeVar, Generic, Optional, Dict, Any

from pykotor.common.misc import CaseInsensitiveDict

from pykotor.common.stream import BinaryReader
from pykotor.extract.capsule import Capsule
from pykotor.extract.file import ResourceIdentifier
from pykotor.extract.installation import Installation, SearchLocation
from pykotor.resource.formats.bwm import read_bwm
from pykotor.resource.formats.gff import read_gff
from pykotor.resource.formats.lyt import LYT
from pykotor.resource.formats.lyt.lyt_auto import read_lyt
from pykotor.resource.formats.mdl import MDL
from pykotor.resource.formats.tpc import read_tpc, TPC
from pykotor.resource.formats.vis import read_vis, VIS
from pykotor.resource.generics.are import ARE, construct_are, read_are
from pykotor.resource.generics.dlg import construct_dlg, read_dlg
from pykotor.resource.generics.git import GIT, construct_git, read_git
from pykotor.resource.generics.ifo import IFO, construct_ifo, read_ifo
from pykotor.resource.generics.pth import construct_pth, read_pth
from pykotor.resource.generics.utc import UTC, construct_utc, read_utc
from pykotor.resource.generics.utd import UTD, construct_utd, read_utd
from pykotor.resource.generics.ute import UTE, construct_ute, read_ute
from pykotor.resource.generics.uti import UTI, construct_uti, read_uti
from pykotor.resource.generics.utm import UTM, construct_utm, read_utm
from pykotor.resource.generics.utp import UTP, construct_utp, read_utp
from pykotor.resource.generics.uts import construct_uts, read_uts
from pykotor.resource.generics.utt import UTT, construct_utt, read_utt
from pykotor.resource.generics.utw import UTW, construct_utw, read_utw
from pykotor.resource.type import ResourceType
from pykotor.tools.model import list_textures

T = TypeVar('T')
SEARCH_ORDER = [SearchLocation.OVERRIDE, SearchLocation.CUSTOM_MODULES, SearchLocation.CHITIN]


class Module:
    def __init__(
            self,
            root: str,
            installation: Installation,
            custom_capsule: Optional[Capsule] = None
    ):
        self._installation = installation
        self._root = root

        self._capsules = [custom_capsule] if custom_capsule is not None else []
        self._capsules.extend([Capsule(installation.module_path() + module) for module in installation.module_names() if root in module])

        for capsule in self._capsules:
            if capsule.exists("module", ResourceType.IFO):
                ifo = read_gff(capsule.resource("module", ResourceType.IFO))
                self._id = ifo.root.get_resref("Mod_Entry_Area").get().lower()
                break
        else:
            raise ValueError("Unable to locate module IFO file for '{}'.".format(root))

        self.resources: CaseInsensitiveDict[ModuleResource] = CaseInsensitiveDict()
        self.reload_resources()

    @staticmethod
    def get_root(
            filepath: str
    ) -> str:
        """
        Returns the root name for a module from the given filepath (or filename). For example "danm13_s.rim" would
        become "danm13".

        Args:
            filepath: The filename or filepath of one of the module encapsulated file.

        Returns:
            The string for the root name of a module.
        """
        root = os.path.basename(filepath).replace(".rim", "").replace(".erf", "").replace(".mod", "").lower()
        roota = root[:5]
        rootb = root[5:]
        if "_" in rootb:
            rootb = rootb[:rootb.index("_")]
        return roota + rootb

    def reload_resources(
            self
    ):
        # Look in module files
        for capsule in self._capsules:
            for resource in capsule:
                resname = resource.resname()
                restype = resource.restype()
                self.add_locations(resname, restype, [capsule.path()])

        # Look for LYT/VIS
        for resource in self._installation.chitin_resources():
            if resource.resname() == self._id:
                self.add_locations(resource.resname(), resource.restype(), [resource.filepath()])

        # Any resource linked in the GIT not present in the module files
        original = self.git().active()
        look_for = []
        for location in self.git().locations():
            self.git().activate(location)
            git = self.git().resource()
            [look_for.append(ResourceIdentifier(creature.resref.get(), ResourceType.UTC)) for creature in git.creatures]
            [look_for.append(ResourceIdentifier(placeable.resref.get(), ResourceType.UTP)) for placeable in
             git.placeables]
            [look_for.append(ResourceIdentifier(door.resref.get(), ResourceType.UTD)) for door in git.doors]
            [look_for.append(ResourceIdentifier(sound.resref.get(), ResourceType.UTS)) for sound in git.sounds]
            [look_for.append(ResourceIdentifier(waypoint.resref.get(), ResourceType.UTW)) for waypoint in git.waypoints]
            [look_for.append(ResourceIdentifier(encounter.resref.get(), ResourceType.UTE)) for encounter in
             git.encounters]
            [look_for.append(ResourceIdentifier(trigger.resref.get(), ResourceType.UTT)) for trigger in git.triggers]
            [look_for.append(ResourceIdentifier(store.resref.get(), ResourceType.UTM)) for store in git.stores]
        self.git().activate(original)

        # Models referenced in LYTs
        original = self.layout().active()
        for location in self.layout().locations():
            self.layout().activate(location)
            layout = self.layout().resource()
            for room in layout.rooms:
                look_for.append(ResourceIdentifier(room.model, ResourceType.MDL))
                look_for.append(ResourceIdentifier(room.model, ResourceType.MDX))
                look_for.append(ResourceIdentifier(room.model, ResourceType.WOK))
        self.layout().activate(original)

        search = self._installation.locations(look_for, [SearchLocation.OVERRIDE, SearchLocation.CHITIN])
        for identifier, locations in search.items():
            self.add_locations(identifier.resname, identifier.restype, [location.filepath for location in locations])

        # Also try get paths for textures in models
        look_for = []
        textures = set()
        for model in self.models():
            with suppress(Exception):
                data = model.data()
                for texture in list_textures(data):
                    textures.add(texture)
        for texture in textures:
            look_for.append(ResourceIdentifier(texture, ResourceType.TPC))
            look_for.append(ResourceIdentifier(texture, ResourceType.TGA))
        search = self._installation.locations(look_for, [SearchLocation.OVERRIDE, SearchLocation.CHITIN,
                                                         SearchLocation.TEXTURES_TPA,
                                                         SearchLocation.TEXTURES_TPB, SearchLocation.TEXTURES_TPC])
        for identifier, locations in search.items():
            if len(locations) == 0:
                continue
            self.add_locations(identifier.resname, identifier.restype, [location.filepath for location in locations])

        for resource in self.resources.values():
            resource.activate()

    def add_locations(
            self,
            resname: str,
            restype: ResourceType,
            locations: List[str]
    ):
        filename = "{}.{}".format(resname, restype)
        if filename in self.resources:
            self.resources[filename].add_locations(locations)
        else:
            self.resources[filename] = ModuleResource(resname, restype, self._installation)
            self.resources[filename].add_locations(locations)

    def installation(
            self
    ) -> Installation:
        return self._installation

    def resource(
            self,
            resname: str,
            restype: ResourceType
    ) -> Optional[ModuleResource]:
        filename = resname + "." + restype.extension
        return self.resources[filename] if filename in self.resources else None

    def layout(
            self
    ) -> ModuleResource[LYT]:
        for filename, resource in self.resources.items():
            if resource.resname().lower() == self._id and resource.restype() == ResourceType.LYT:
                return resource

    def vis(
            self
    ) -> ModuleResource[VIS]:
        for filename, resource in self.resources.items():
            if resource.resname().lower() == self._id and resource.restype() == ResourceType.VIS:
                return resource

    def are(
            self
    ) -> ModuleResource[ARE]:
        for filename, resource in self.resources.items():
            if resource.resname().lower() == self._id and resource.restype() == ResourceType.ARE:
                return resource

    def git(
            self
    ) -> ModuleResource[GIT]:
        for filename, resource in self.resources.items():
            if resource.resname().lower() == self._id and resource.restype() == ResourceType.GIT:
                return resource

    def info(
            self
    ) -> ModuleResource[IFO]:
        for filename, resource in self.resources.items():
            if resource.resname().lower() == "module" and resource.restype() == ResourceType.IFO:
                return resource

    def creature(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTC]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTC:
                return resource
        return None

    def creatures(
            self
    ) -> List[ModuleResource[UTC]]:
        creatures = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTC:
                creatures.append(resource)
        return creatures

    def placeable(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTP]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTP:
                return resource
        return None

    def placeables(
            self
    ) -> List[ModuleResource[UTP]]:
        placeables = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTP:
                placeables.append(resource)
        return placeables

    def door(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTD]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTD:
                return resource
        return None

    def doors(
            self
    ) -> List[ModuleResource[UTD]]:
        doors = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTD:
                doors.append(resource)
        return doors

    def item(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTI]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTI:
                return resource
        return None

    def items(
            self
    ) -> List[ModuleResource[UTI]]:
        doors = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTD:
                doors.append(resource)
        return doors

    def encounter(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTE]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTE:
                return resource
        return None

    def encounters(
            self
    ) -> List[ModuleResource[UTE]]:
        encounters = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTE:
                encounters.append(resource)
        return encounters

    def store(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTM]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTM:
                return resource
        return None

    def stores(
            self
    ) -> List[ModuleResource[UTM]]:
        stores = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTM:
                stores.append(resource)
        return stores

    def trigger(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTT]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTT:
                return resource
        return None

    def triggers(
            self
    ) -> List[ModuleResource[UTT]]:
        triggers = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTT:
                triggers.append(resource)
        return triggers

    def waypoint(
            self,
            resname: str
    ) -> Optional[ModuleResource[UTW]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.UTW:
                return resource
        return None

    def waypoints(
            self
    ) -> List[ModuleResource[UTW]]:
        waypoints = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.UTW:
                waypoints.append(resource)
        return waypoints

    def model(
            self,
            resname: str
    ) -> Optional[ModuleResource[MDL]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.MDL:
                return resource
        return None

    def model_ext(
            self,
            resname: str
    ) -> Optional[ModuleResource]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() == ResourceType.MDX:
                return resource
        return None

    def models(
            self
    ) -> List[ModuleResource[MDL]]:
        models = []
        for resource in self.resources.values():
            if resource.restype() == ResourceType.MDL:
                models.append(resource)
        return models

    def texture(
            self,
            resname: str
    ) -> Optional[ModuleResource[TPC]]:
        for resource in self.resources.values():
            if resname == resource.resname() and resource.restype() in [ResourceType.TPC, ResourceType.TGA]:
                return resource
        return None

    def textures(
            self
    ) -> List[ModuleResource[MDL]]:
        textures = []
        for resource in self.resources.values():
            if resource.restype() in [ResourceType.TPC, ResourceType.TGA]:
                textures.append(resource)
        return textures


class ModuleResource(Generic[T]):
    def __init__(
            self,
            resname: str,
            restype: ResourceType,
            installation: Installation
    ):
        self._resname: str = resname
        self._installation = installation
        self._restype: ResourceType = restype
        self._active: Optional[str] = None
        self._resource: Any = None
        self._locations: List[str] = []

    def resname(
            self
    ) -> str:
        """
        Returns the resource name.

        Returns:
            The resource name.
        """
        return self._resname

    def restype(
            self
    ) -> ResourceType:
        """
        Returns the type of resource stored.

        Returns:
            The resource type.
        """
        return self._restype

    def localized_name(
            self
    ) -> Optional[str]:
        res = self.resource()
        if res is None:
            return None
        elif isinstance(res, UTC):
            return self._installation.string(res.first_name) + " " + self._installation.string(res.last_name)
        elif isinstance(res, UTP):
            return self._installation.string(res.name)
        elif isinstance(res, UTD):
            return self._installation.string(res.name)
        elif isinstance(res, UTW):
            return self._installation.string(res.name)
        elif isinstance(res, UTT):
            return self._installation.string(res.name)
        elif isinstance(res, UTE):
            return self._installation.string(res.name)
        elif isinstance(res, UTM):
            return self._installation.string(res.name)
        else:
            return None

    def data(
            self
    ) -> bytes:
        """
        Opens the file at the active location and returns the data.

        Raises:
            ValueError: If no file is active.

        Returns:
            The bytes data of the active file.
        """

        if self._active is None:
            raise ValueError("No file is currently active for resource '{}.{}'.".format(self.resname, self._restype.extension))
        elif self._active.endswith(".erf") or self._active.endswith(".mod") or self._active.endswith(".rim"):
            capsule = Capsule(self._active)
            return capsule.resource(self._resname, self._restype)
        elif self._active.endswith("bif"):
            return self._installation.resource(self._resname, self._restype, [SearchLocation.CHITIN]).data
        else:
            return BinaryReader.load_file(self._active)

    def resource(
            self
    ) -> Optional[T]:
        """
        Returns the cached resource object. If no object has been cached, then it will load the object.

        Returns:
            The resource object.
        """

        if self._resource is None:
            conversions = {
                ResourceType.UTC: (lambda data: read_utc(data)),
                ResourceType.UTP: (lambda data: read_utp(data)),
                ResourceType.UTD: (lambda data: read_utd(data)),
                ResourceType.UTI: (lambda data: read_uti(data)),
                ResourceType.UTM: (lambda data: read_utm(data)),
                ResourceType.UTE: (lambda data: read_ute(data)),
                ResourceType.UTT: (lambda data: read_utt(data)),
                ResourceType.UTW: (lambda data: read_utw(data)),
                ResourceType.UTS: (lambda data: read_uts(data)),
                ResourceType.DLG: (lambda data: read_dlg(data)),
                ResourceType.PTH: (lambda data: read_pth(data)),
                ResourceType.NCS: (lambda data: data),
                ResourceType.TPC: (lambda data: read_tpc(data)),
                ResourceType.LYT: (lambda data: read_lyt(data)),
                ResourceType.VIS: (lambda data: read_vis(data)),
                ResourceType.IFO: (lambda data: read_ifo(data)),
                ResourceType.ARE: (lambda data: read_are(data)),
                ResourceType.GIT: (lambda data: read_git(data)),
                ResourceType.WOK: (lambda data: read_bwm(data))
            }

            if self._active is None:
                self._resource = None
            elif self._active.endswith(".erf") or self._active.endswith(".mod") or self._active.endswith(".rim"):
                data = Capsule(self._active).resource(self._resname, self._restype)
                self._resource = conversions[self._restype](data)
            elif self._active.endswith("bif"):
                data = self._installation.resource(self._resname, self._restype, [SearchLocation.CHITIN]).data
                self._resource = conversions[self._restype](data)
            else:
                data = BinaryReader.load_file(self._active)
                self._resource = conversions[self._restype](data)

        return self._resource

    def add_locations(
            self,
            filepaths: List[str]
    ) -> None:
        """
        Adds a list of filepaths to the list of locations stored for the resource. If a filepath already exists, it is
        ignored.

        Args:
            filepaths: A list of filepaths pointing to a location for the resource.
        """
        self._locations.extend([filepath for filepath in filepaths if filepath not in self._locations])
        if self._active is None and self._locations:
            self.activate(self._locations[0])

    def locations(
            self,
    ) -> List[str]:
        return self._locations

    def activate(
            self,
            filepath: str = None
    ) -> None:
        """
        Sets the active file to the specified path. Calling this method will reset the loaded resource.

        Raises:
            ValueError: If the filepath is not stored in the resource list of locations.

        Args:
            filepath: The new active file.
        """
        self._resource = None
        if filepath is None:
            self._active = self._locations[0] if len(self._locations) > 0 else None
        elif filepath in self._locations:
            self._active = filepath
        else:
            raise ValueError("The filepath '{}' is not being tracked as a location for the resource.".format(self._active))

    def unload(
            self
    ) -> None:
        """
        Clears the cached resource object from memory.
        """
        self._resource = None

    def reload(
            self
    ) -> None:
        """
        Reloads the resource object from the active location.
        """
        self._resource = None
        self.resource()

    def active(
            self
    ) -> Optional[str]:
        """
        Returns the filepath of the currently active file for the resource.

        Returns:
            Filepath to the active resource.
        """
        return self._active



from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.mdl import MDL, MDLBinaryReader, MDLBinaryWriter, MDLAsciiReader, MDLAsciiWriter
from pykotor.resource.type import FileFormat, SOURCE_TYPES, TARGET_TYPES


def detect_mdl(source: SOURCE_TYPES, offset: int = 0) -> FileFormat:
    """
    Returns what format the MDL data is believed to be in. This function performs a basic check and does not guarantee
    accuracy of the result or integrity of the data.

    Args:
        source: Source of the MDL data.
        offset: Offset into the source data.

    Returns:
        The format of the MDL data.
    """
    try:
        if isinstance(source, str):
            with BinaryReader.from_file(source, offset) as reader:
                first4 = reader.read_bytes(4)
                file_format = FileFormat.BINARY if first4 == b'\x00\x00\x00\x00' else FileFormat.ASCII
        elif isinstance(source, bytes) or isinstance(source, bytearray):
            file_format = FileFormat.BINARY if source[:4] == b'\x00\x00\x00\x00' else FileFormat.ASCII
        elif isinstance(source, BinaryReader):
            first4 = source.read_bytes(4)
            file_format = FileFormat.BINARY if first4 == b'\x00\x00\x00\x00' else FileFormat.ASCII
            source.skip(-4)
        else:
            file_format = FileFormat.INVALID
    except IOError:
        file_format = FileFormat.INVALID

    return file_format


def load_mdl(source: SOURCE_TYPES, offset: int = 0, source_ext: SOURCE_TYPES = None, offset_ext: int = 0) -> MDL:
    """
    Returns an MDL instance from the source. The file format (binary or ascii) is automatically determined before
    parsing the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.
        source_ext: Source of the MDX data, if available.
        offset_ext: Offset into the source_ext data.

    Raises:
        ValueError: If the file was corrupted or in an unsupported format.

    Returns:
        An MDL instance.
    """
    file_format = detect_mdl(source, offset)

    try:
        if file_format == FileFormat.BINARY:
            return MDLBinaryReader(source, offset).load()
        elif file_format == FileFormat.ASCII:
            return MDLAsciiReader(source).load()
        else:
            raise ValueError
    except (IOError, ValueError):
        raise ValueError("Tried to load an unsupported or corrupted MDL file.")


def write_mdl(mdl: MDL, target: TARGET_TYPES, file_format: FileFormat = FileFormat.BINARY,
              target_ext: TARGET_TYPES = None) -> None:
    """
    Writes the MDL data to the target location with the specified format (binary or ascii).

    Args:
        mdl: The MDL file being written.
        target: The location to write the data to.
        file_format: The file format.
        target_ext: The location to write the MDX data to (if binary).

    Raises:
        ValueError: If an unsupported FileFormat is passed.
    """
    if file_format == FileFormat.BINARY:
        MDLBinaryWriter(mdl, target).write()
    elif file_format == FileFormat.ASCII:
        MDLAsciiWriter(mdl, target).write()
    else:
        raise ValueError("Unsupported format specified; use BINARY or ASCII.")



from __future__ import annotations

from typing import Optional

from pykotor.resource.formats.rim import RIM
from pykotor.resource.type import ResourceType, SOURCE_TYPES, TARGET_TYPES, ResourceReader, ResourceWriter


class RIMBinaryReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._rim: Optional[RIM] = None

    def load(
            self,
            auto_close: bool = True
    ) -> RIM:
        self._rim = RIM()

        file_type = self._reader.read_string(4)
        file_version = self._reader.read_string(4)

        if file_type != "RIM ":
            raise TypeError("The RIM file type that was loaded was unrecognized.")

        if file_version != "V1.0":
            raise TypeError("The RIM version that was loaded is not supported.")

        self._reader.skip(4)
        entry_count = self._reader.read_uint32()
        offset_to_keys = self._reader.read_uint32()

        resrefs = []
        resids = []
        restypes = []
        resoffsets = []
        ressizes = []
        self._reader.seek(offset_to_keys)
        for i in range(entry_count):
            resrefs.append(self._reader.read_string(16))
            restypes.append(self._reader.read_uint32())
            resids.append(self._reader.read_uint32())
            resoffsets.append(self._reader.read_uint32())
            ressizes.append(self._reader.read_uint32())

        for i in range(entry_count):
            self._reader.seek(resoffsets[i])
            resdata = self._reader.read_bytes(ressizes[i])
            self._rim.set(resrefs[i], ResourceType.from_id(restypes[i]), resdata)

        if auto_close:
            self._reader.close()

        return self._rim


class RIMBinaryWriter(ResourceWriter):
    FILE_HEADER_SIZE = 120
    KEY_ELEMENT_SIZE = 32

    def __init__(
            self,
            rim: RIM,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._rim = rim

    def write(
            self,
            auto_close: bool = True
    ) -> None:
        entry_count = len(self._rim)
        offset_to_keys = RIMBinaryWriter.FILE_HEADER_SIZE

        self._writer.write_string("RIM ")
        self._writer.write_string("V1.0")
        self._writer.write_uint32(0)
        self._writer.write_uint32(entry_count)
        self._writer.write_uint32(offset_to_keys)
        self._writer.write_bytes(b'\0' * 100)

        resid = 0
        data_offset = offset_to_keys + RIMBinaryWriter.KEY_ELEMENT_SIZE * entry_count
        for resource in self._rim:
            self._writer.write_string(resource.resref.get(), string_length=16)
            self._writer.write_uint32(resource.restype.type_id)
            self._writer.write_uint32(resid)
            self._writer.write_uint32(data_offset)
            self._writer.write_uint32(len(resource.data))
            resid += 1
            data_offset += len(resource.data)

        for resource in self._rim:
            self._writer.write_bytes(resource.data)

        if auto_close:
            self._writer.close()



'''OpenGL extension VERSION.EGL_1_5

This module customises the behaviour of the 
OpenGL.raw.EGL.VERSION.EGL_1_5 to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/VERSION/EGL_1_5.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.EGL import _types, _glgets
from OpenGL.raw.EGL.VERSION.EGL_1_5 import *
from OpenGL.raw.EGL.VERSION.EGL_1_5 import _EXTENSION_NAME

def glInitEgl15VERSION():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )


### END AUTOGENERATED SECTION
from ..EXT.platform_base import eglGetPlatformDisplayEXT
from OpenGL.extensions import alternate as _alternate
eglGetPlatformDisplay.extension = None
eglGetPlatformDisplay.force_extension = True 

eglGetPlatformDisplay = _alternate(
    'eglGetPlatformDisplay',
    eglGetPlatformDisplayEXT,
    eglGetPlatformDisplay,
)


from unittest import TestCase

from pykotor.common.language import LocalizedString
from pykotor.common.misc import EquipmentSlot, Game
from pykotor.resource.formats.gff import load_gff
from pykotor.resource.generics.are import construct_are, dismantle_are

TEST_FILE = "../../files/test.are"


class TestARE(TestCase):
    def test_change_texture(self):
        ...




from typing import List, Optional, Dict

from pykotor.common.stream import BinaryReader
from pykotor.extract.file import FileResource, ResourceResult, ResourceIdentifier
from pykotor.resource.type import ResourceType


class Capsule:
    """
    Chitin object is used for loading the list of resources stored in the .erf/.rim/.mod files used by the game.
    Resource data is not actually stored in memory by default but is instead loaded up on demand with the
    Capsule.resource() method.

    Capsule data is read-only, use ERF or RIM classes instead for creating and editing files.
    """
    def __init__(
            self,
            path: str
    ):
        self._path: str = path
        self._resources: List[FileResource] = []

        self.reload()

    def __iter__(
            self
    ):
        for resource in self._resources:
            yield resource

    def __len__(
            self
    ):
        return len(self._resources)

    def resource(
            self,
            resref: str,
            restype: ResourceType,
            reload: bool = False
    ) -> Optional[bytes]:
        """
        Returns the bytes data of the specified resource. If the resource does not exist then returns None instead.

        Args:
            resref: The resource ResRef.
            restype: The resource type.
            reload: If True Capsule will reload the ERF before opening rather than using cached offsets.

        Returns:
            None or bytes data of resource.
        """
        if reload:
            self.reload()

        query = ResourceIdentifier(resref, restype)
        resource = next((resource for resource in self._resources if resource == query), None)
        return None if resource is None else resource.data()

    def batch(
            self,
            queries: List[ResourceIdentifier],
            reload: bool = False
    ) -> Dict[ResourceIdentifier, Optional[ResourceResult]]:
        if reload:
            self.reload()

        results = {}
        reader = BinaryReader.from_file(self.path())

        for query in queries:
            results[query] = None
            if self.exists(query.resname, query.restype):
                resource = next((resource for resource in self._resources if resource == query), None)
                if resource is not None:
                    reader.seek(resource.offset())
                    data = reader.read_bytes(resource.size())
                    results[query] = ResourceResult(query.resname, query.restype, self.path(), data)

        reader.close()
        return results

    def exists(
            self,
            resref: str,
            restype: ResourceType,
            reload: bool = False
    ) -> bool:
        if reload:
            self.reload()

        query = ResourceIdentifier(resref, restype)
        resource = next((resource for resource in self._resources if resource == query), None)
        return resource is not None

    def info(
            self,
            resref: str,
            restype: ResourceType,
            reload: bool = False
    ) -> FileResource:
        if reload:
            self.reload()

        query = ResourceIdentifier(resref, restype)
        resource = next((resource for resource in self._resources if resource == query), None)
        return resource

    def reload(
            self
    ):
        """
        Reload the list of resource info linked from the module file.
        """
        with BinaryReader.from_file(self._path) as reader:
            file_type = reader.read_string(4)
            file_version = reader.read_string(4)

            if file_type in ("ERF ", "MOD "):
                self._load_erf(reader)
            elif file_type == "RIM ":
                self._load_rim(reader)
            else:
                raise ValueError("File '{}' was not an ERF/MOD/RIM.".format(self._path))

    def path(
            self
    ) -> str:
        return self._path

    def _load_erf(
            self,
            reader: BinaryReader
    ):
        reader.skip(8)
        entry_count = reader.read_uint32()
        reader.skip(4)
        offset_to_keys = reader.read_uint32()
        offset_to_resources = reader.read_uint32()

        resrefs = []
        resids = []
        restypes = []
        reader.seek(offset_to_keys)
        for i in range(entry_count):
            resrefs.append(reader.read_string(16))
            resids.append(reader.read_uint32())
            restypes.append(ResourceType.from_id(reader.read_uint16()))
            reader.skip(2)

        reader.seek(offset_to_resources)
        for i in range(entry_count):
            res_offset = reader.read_uint32()
            res_size = reader.read_uint32()
            self._resources.append(FileResource(resrefs[i], restypes[i], res_size, res_offset, self._path))

    def _load_rim(
            self,
            reader: BinaryReader
    ):
        reader.skip(4)
        entry_count = reader.read_uint32()
        offset_to_entries = reader.read_uint32()

        reader.seek(offset_to_entries)
        for i in range(entry_count):
            resref = reader.read_string(16)
            restype = ResourceType.from_id(reader.read_uint32())
            res_id = reader.read_uint32()
            offset = reader.read_uint32()
            size = reader.read_uint32()
            self._resources.append(FileResource(resref, restype, size, offset, self._path))




from unittest import TestCase

from pykotor.resource.formats.lip import LIP, LIPShape, LIPBinaryReader, detect_lip, write_lip, LIPXMLReader
from pykotor.resource.type import FileFormat

BINARY_TEST_FILE = "../../files/test.lip"
XML_TEST_FILE = "../../files/test.lip.xml"


class TestLIP(TestCase):
    def test_binary_io(self):
        self.assertEqual(detect_lip(BINARY_TEST_FILE), FileFormat.BINARY)

        lip = LIPBinaryReader(BINARY_TEST_FILE).load()
        self.validate_io(lip)

        data = bytearray()
        write_lip(lip, data, FileFormat.BINARY)
        lip = LIPBinaryReader(data).load()
        self.validate_io(lip)

    def test_xml_io(self):
        self.assertEqual(detect_lip(XML_TEST_FILE), FileFormat.XML)

        lip = LIPXMLReader(XML_TEST_FILE).load()
        self.validate_io(lip)

        data = bytearray()
        write_lip(lip, data, FileFormat.XML)
        lip = LIPXMLReader(data).load()
        self.validate_io(lip)

    def validate_io(self, lip: LIP):
        self.assertAlmostEqual(lip.length, 1.50, 3)
        self.assertEqual(LIPShape.EE, lip.get(0).shape)
        self.assertEqual(LIPShape.OOH, lip.get(1).shape)
        self.assertEqual(LIPShape.TH, lip.get(2).shape)
        self.assertAlmostEqual(0.0, lip.get(0).time, 4)
        self.assertAlmostEqual(0.7777, lip.get(1).time, 4)
        self.assertAlmostEqual(1.25, lip.get(2).time, 4)



import pathlib
from setuptools import setup, find_packages, find_namespace_packages

HERE = pathlib.Path(__file__).parent

NAME = "PyKotorGL"
VERSION = "1.1.2"
AUTHOR = "Nicholas Hugi"
DESCRIPTION = "Render modules from both KotOR games."
PACKAGES = find_namespace_packages()
URL = "https://github.com/NickHugi/PyKotorGL"

README = (HERE / "README.rd").read_text()
REQUIREMENTS = (HERE / "requirements.txt").read_text()

setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    description=DESCRIPTION,
    install_requires=REQUIREMENTS,
    long_description=README,
    packages=PACKAGES,
    url=URL
)
